//ROS Libraries
#include "ros.h"
#include <math.h>
#include <geometry_msgs/Twist.h>
#include <ros/time.h>
#include <tf/tf.h>
#include <tf/transform_broadcaster.h>
#include <nav_msgs/Odometry.h>
#include <std_msgs/String.h>
#include <geometry_msgs/Vector3Stamped.h>
#include <geometry_msgs/Vector3.h>

#include <HardwareSerial.h>
//ODrive Library
#include <ODriveArduino.h>

//ODrive Objects
ODriveArduino odrive1(Serial2);

// Define the robot parameters
#define romr_width (300) //mm
#define wheel_diameter (165.1) //mm 
#define wheel_circumference (518.362788)//mm
#define encoder_cpr (90) //count per revolution
#define rev (5) //motor revolutions per wheel circunfrence

// ROS handle
ros::NodeHandle nh;
geometry_msgs::TransformStamped t;
nav_msgs::Odometry odom_msg;
ros::Publisher odom_pub("odom", &odom_msg);
tf::TransformBroadcaster broadcaster;

// tf variables to be broadcast
double x = 0;
double y = 0;
double theta = 0;

char base_link[] = "/base_link";
char odom[] = "/odom";

// cmd_vel variables to drive the robot with
float request_x;
float request_z;

// set timers for the sub_main loop
unsigned long currentMillis;
long previousMillis = 0; // set up timers
float loopTime = 10;

int requested_state;
float modifier_linear;
float modifier_angular;

// convert the parameter units into turn/s
float conversion_linear;
float conversion_angular;
int m0_linear; 
int m1_linear; 
int m0_angular;        
int m1_angular; 

// get positions and velocities from ODrive

long pos0;                   
long pos1;   

// work out the difference on each loop, and bookmark the old value
long pos0_old;
long pos0_diff;
long pos1_diff;
long pos1_old;
float pos0_mm_diff;
float pos1_mm_diff;
float pos_average_mm_diff;
float pos_total_mm;


void velCallback( const geometry_msgs::Twist& vel)
{
request_x = vel.linear.x;
request_z = vel.angular.z;

//request_x = 0.02*vel.linear.x;  //decrease by 98%
//request_z = 0.02*vel.angular.z; //decrease by 98%
}
// subscribe to the cmd_vel
ros::Subscriber<geometry_msgs::Twist> sub("cmd_vel" , velCallback);

//  Setup 

void setup() {

nh.getHardware()->setBaud(115200); // set baud rate to 115200
nh.initNode(); // init ROS
nh.subscribe(sub); // subscribe to cmd_vel
nh.advertise(odom_pub);
broadcaster.init(nh); // set up broadcaster

Serial.begin(115200); // ROS serial communication
Serial2.begin(115200); // ODrive serial communication

}

// Main loop

void loop() {

nh.spinOnce(); // listen to ROS messages and activate the callback if necessary

currentMillis = millis();
      if (currentMillis - previousMillis >= loopTime) { // run a loop every 10ms
      previousMillis = currentMillis; // reset the clock
      
      // set the ODrive controller to closed loop state
        requested_state = AXIS_STATE_CLOSED_LOOP_CONTROL;
        odrive1.run_state(0, requested_state, false); // don't wait
        odrive1.run_state(1, requested_state, false); // don't wait
  

// scaling factor because the wheels are squashy / there is wheel slip etc.
  modifier_linear = -1.03;
  modifier_angular = -0.92;

// convert the parameter units into turn/s
  conversion_linear = (rev * 1000)/wheel_circumference;
  conversion_angular = ((romr_width/2) * rev)/wheel_circumference;
  m0_linear = request_x * conversion_linear * modifier_linear; 
  m1_linear = request_x * conversion_linear * modifier_linear; 
  m0_angular = request_z * conversion_angular * modifier_angular;        
  m1_angular = request_z * conversion_angular * modifier_angular;   

// reverse the direction of rotation of one of the wheel       
m1_linear = m1_linear*-1;

// set the velocity control
odrive1.SetVelocity(0, m0_linear + m0_angular); 
odrive1.SetVelocity(1, m1_linear + m1_angular);
 
// get positions and velocities from ODrive

 pos0 = (odrive1.GetPosition(1)) *-1;                   
 pos1 = odrive1.GetPosition(0);   

pos0_diff = pos0 - pos0_old;
pos1_diff = pos1 - pos1_old;            
pos0_old = pos0;
pos1_old = pos1;
    
// calc mm from encoder counts
pos0_mm_diff = pos0_diff /1; //9.64;     //83.44;
pos1_mm_diff = pos1_diff /1; //9.64;     //83.44;

// calc distance travelled based on average of both wheels
pos_average_mm_diff = (pos0_mm_diff + pos1_mm_diff) / 2;   // difference in each cycle
pos_total_mm += pos_average_mm_diff;                       // calc total running total distance

// calc angle or rotation to broadcast with tf
theta += (pos1_mm_diff - pos0_mm_diff) / romr_width;  //360;
            
if (theta > PI)
theta -= TWO_PI;
if (theta < (-PI))
theta += TWO_PI;

 // calc x and y to broadcast with tf

y += pos_average_mm_diff * sin(theta);
x += pos_average_mm_diff * cos(theta);

// *** broadcast odom->base_link transform with tf ***
//first, we'll publish the transform over tf
geometry_msgs::TransformStamped t;                              
t.header.stamp = nh.now();
t.header.frame_id = odom;
t.child_frame_id = base_link;            
t.transform.translation.x = x/1000;   // convert to metres
t.transform.translation.y = y/1000;
t.transform.translation.z = 0;
 //since all odometry is 6DOF we'll need a quaternion created from yaw           
t.transform.rotation = tf::createQuaternionFromYaw(theta);

 //send the transform                    
broadcaster.sendTransform(t);

// *** broadcast odom message ***
 //next, we'll publish the odometry message over ROS
nav_msgs::Odometry odom_msg;
odom_msg.header.stamp = nh.now();
odom_msg.header.frame_id = odom;
 //set the position
odom_msg.pose.pose.position.x = x/1000;
odom_msg.pose.pose.position.y = y/1000;
odom_msg.pose.pose.position.z = 0.0;
odom_msg.pose.pose.orientation = tf::createQuaternionFromYaw(theta);

//set the velocity
odom_msg.child_frame_id = base_link;
odom_msg.twist.twist.linear.x = ((pos0_mm_diff + pos1_mm_diff) / 2)/10;     // forward linear velovity
odom_msg.twist.twist.linear.y = 0.0;                                         // robot does not move sideways
odom_msg.twist.twist.angular.z = ((pos1_mm_diff - pos0_mm_diff) / romr_width)*100;  // anglular velocity

//publish the message
odom_pub.publish(&odom_msg);
//nh.spinOnce();

 }

}
